"""
Simple parser for the Klog plaintext format, generates a high level object of a Klog file.
Is not optimized to be fast, but since the format is simple, it should be fast enough.
"""

import datetime
import re
from dataclasses import asdict
from functools import cache

from parsy import Parser, Result, fail, generate, regex, seq, string, whitespace

from .syntax import Duration, Entry, Range, Record, ShouldTotal

TAG_REGEX = re.compile(r'#\S+')
EOL = regex(r'\s*\n')


@cache
def c_regex(exp, flags=0, group=0) -> Parser:
    """
    Memoized version of ``parsy.regex``.
    Can be used for caching expensive regexes, saves some compiling time.

    The difference in cost is ultimately negligible, since compiling on the fly
    is quick, but it shaves off a little time.

    :return: parsy.Parser object.
    """
    if isinstance(exp, (str, bytes)):
        exp = re.compile(exp, flags)
    if isinstance(group, (str, int)):
        group = (group,)

    @Parser
    def regex_parser(stream, index):
        match = exp.match(stream, index)
        if match:
            return Result.success(match.end(), match.group(*group))
        else:
            return Result.failure(index, exp.pattern)

    return regex_parser


@generate
def date():
    y = yield c_regex(r'\d{4}').map(int)
    sep = yield string('-') | string('/')
    m = yield c_regex(r'\d{2}').map(int)
    yield string(sep)
    d = yield c_regex(r'\d{2}').map(int)
    return datetime.date(y, m, d)


@generate
def duration():
    sign = yield (string('-') | string('+')).optional()

    hours = yield c_regex(r'(\d+)h', group=1).optional()
    minutes = yield c_regex(r'(\d+)m', group=1).optional()

    if hours is None and minutes is None:
        return fail('Duration value should have a minutes or hours value.')

    return Duration(
        is_neg=sign is not None and sign == '-',
        hours=0 if hours is None else hours,
        minutes=0 if minutes is None else minutes,
    )


@generate
def should_total():
    d = yield duration << string('!')
    return ShouldTotal(**asdict(d))


# TODO: dont allow duplicates
properties = should_total


@generate
def time():
    h = yield c_regex(r'\d{1,2}').map(int)
    yield string(':')
    m = yield c_regex(r'\d{1,2}').map(int)
    suffix = yield regex(r'(:?a|p)m').optional()

    is_pm = suffix is not None and suffix == 'pm'

    # Convert 12 hour clock to 24 hour if suffix is set
    if is_pm and h != 12:
        h += 12
    elif not is_pm and h == 12:
        h = 0

    return datetime.time(hour=h, minute=m)


@generate
def time_range():
    start_time = seq(
        string('<').optional().map(lambda x: x is not None),
        time
    )

    start = yield start_time | c_regex(r'\?+').result(None)

    yield c_regex(r'\s*-\s*')

    end_time = seq(
        time,
        (string('>').optional().map(lambda x: x is not None))
    ).map(lambda t: (t[1], t[0]))

    end = yield end_time | c_regex(r'\?+').result(None)

    # Check that, if start and end are time values, that they are in the correct order
    if (
        isinstance(start, tuple)
        and isinstance(end, tuple)
        and start[1] > end[1]
    ):
        return fail('Time in range must be in chronological order')

    return Range(
        start=start,
        end=end,
    )


@generate
def entry():
    # Indent
    yield c_regex(r'(( {3,4})|\t)')

    t = yield time_range | duration

    description = yield c_regex(r' *') >> c_regex('.*')
    yield EOL

    return Entry(
        time=t,
        description=description,
    )


@generate
def record():
    d = yield date

    prop = yield (
        whitespace
        >> string('(')
        >> properties
        << string(')')
    ).optional()

    yield EOL

    summary = yield c_regex(r'(\w.*)\n', group=1).many()
    entries = yield entry.many()

    tags = TAG_REGEX.findall(' '.join(summary))

    return Record(
        date=d,
        properties=[prop],
        summary=summary,
        entries=entries,
        tags=tags,
    )


def parse(inp: str) -> list[Record]:
    padded_record = EOL.many() >> record << EOL.many()
    records = padded_record.many().parse(inp)
    return records



from datetime import datetime

import click
import parsy

import klogpy.parser as parser
import klogpy.syntax as syntax
from klogpy import __version__
from klogpy.config import CONFIG_DIR, RECORD_STORE, get_local_config


@click.command('init')
def _init():
    """Initialize a new record store if it does not exist"""
    created, _ = get_local_config(True)

    if created:
        click.secho(
            f'Created a new record store in {CONFIG_DIR.resolve()}', fg='green'
        )
    else:
        click.echo(f'Record store in {CONFIG_DIR.resolve()} already exists')


@click.command('entry', short_help='Manipulate entries for the active day')
@click.pass_context
@click.option('-m', '--message', help='Add a summary to the record')
@click.option('-l', '--list', 'should_list', is_flag=True, help='Print the entries of the currently selected record')
@click.argument('val', type=str, required=False, metavar='[ENTRY]')
def _entry(ctx, message, should_list, val):
    """Manipulate entries for the active day

    \b
    VAL must be a time or range entry
    Some examples of entries:

    \b
    - "2h30m"
    - "10:00am - 12:00pm"
    """
    if should_list:
        _, conf = get_local_config()

        if conf is None:
            click.secho(
                f'Record store ({RECORD_STORE}) does not exist, or file is corrupt',
                fg='red'
            )
            ctx.exit()

        if conf.current_record is None:
            click.echo('Currently no pending records')
            ctx.exit()

        for e in conf.current_record.entries:
            click.echo(e.serialize())

        ctx.exit()

    if val is None:
        click.echo(ctx.get_help(), color=ctx.color)
        ctx.exit()

    try:
        time_val = (parser.time_range | parser.duration).parse(val)
    except parsy.ParseError:
        click.secho(
            'Given entry value does not match duration or time range', fg='red')
        return

    e = syntax.Entry(
        time=time_val,
        description=message,
    )

    created, conf = get_local_config(should_create=True)

    if created:
        click.secho(
            f'Created a new record store in {CONFIG_DIR.resolve()}', fg='green')
        return

    conf.current_record.entries.append(e)
    conf.commit()

    click.echo(f'Added {e.serialize()}')


@click.group('record', invoke_without_command=True)
@click.pass_context
@click.option('-l', '--list', 'should_list', is_flag=True, help='List all records in the record store')
def _record(ctx, should_list: bool):
    """Create or modify records"""
    if should_list:
        _, conf = get_local_config()

        if conf is None:
            click.secho(
                f'Record store ({RECORD_STORE}) does not exist, or file is corrupt',
                fg='red'
            )
            return

        for i, record in enumerate(conf.records):
            if i == conf.current_record_index:
                click.secho('* ', nl=False, fg='green')
            else:
                click.echo('  ', nl=False)

            click.echo('  '.join(record.serialize().split('\n')))

        return

    if ctx.invoked_subcommand is None:
        click.echo(ctx.get_help(), color=ctx.color)
        ctx.exit()


@click.command('finalize')
@click.option('-m', '--message', multiple=True)
@click.option('-e', '--edit', is_flag=True)
def _finalize(message, edit):
    """Write the current selected record to the Klog file"""
    _, conf = get_local_config()

    if conf is None:
        click.secho(
            f'Record store ({RECORD_STORE}) does not exist, or file is corrupt', fg='red')
        return

    conf.current_record.summary = [m.strip() for m in message]

    if edit:
        # Add previous message in text buffer if present
        text = click.edit('\n'.join(conf.current_record.summary))

        # Remove empty lines
        conf.current_record.summary = [
            line for line in text.split('\n') if line != '']

    try:
        pushed = conf.write_selected()
    except Exception as e:
        click.secho(str(e), fg='red')
    else:
        click.secho('Successfully added record\n', fg='green')
        click.echo(pushed.serialize())


@click.command('new')
def _new():
    """Create a new record for today"""
    rec = syntax.Record(datetime.today())
    _, conf = get_local_config()

    if conf is None:
        click.secho(
            f'Record store ({RECORD_STORE}) does not exist, or file is corrupt', fg='red')
        return

    conf.push_record(rec)
    conf.commit()


_record.add_command(_finalize)
_record.add_command(_new)


@click.group()
@click.version_option(__version__)
def klg():
    pass


klg.add_command(_init)
klg.add_command(_entry)
klg.add_command(_record)



import datetime
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from functools import cached_property
from typing import Optional, Union


class Serializable(ABC):
    @abstractmethod
    def serialize(self) -> str:
        pass


class Time(ABC):
    @property
    @abstractmethod
    def total_minutes(self) -> Optional[int]:
        pass


@dataclass
class Duration(Time, Serializable):
    is_neg: bool
    hours: int
    minutes: int

    @cached_property
    def total_minutes(self) -> int:
        total_minutes = self.hours * 60 + self.minutes
        return -total_minutes if self.is_neg else total_minutes

    def serialize(self) -> str:
        builder = []

        if self.is_neg:
            builder.append('-')

        builder.append(f'{self.hours}h{self.minutes}m')

        return ''.join(builder)


@dataclass
class Range(Time, Serializable):
    start: Optional[tuple[bool, datetime.time]]
    end: Optional[tuple[bool, datetime.time]]

    @cached_property
    def total_minutes(self) -> Optional[int]:
        # TODO
        return None

    def serialize(self) -> str:
        builder = []

        if self.start is not None:
            o, t = self.start
            builder.append(f'{"<" if o else ""}{t.hour:02}:{t.minute:02}')
        else:
            builder.append('?')

        builder.append(' - ')

        if self.end is not None:
            o, t = self.end
            builder.append(f'{t.hour:02}:{t.minute:02}{">" if o else ""}')
        else:
            builder.append('?')

        return ''.join(builder)


@dataclass
class Entry(Serializable):
    time: Union[Duration, Range]
    description: Optional[str]

    def serialize(self) -> str:
        s = self.time.serialize()
        if self.description is not None:
            s += f' {self.description}'

        return s


class ShouldTotal(Duration):
    pass


Property = Union[ShouldTotal]


@dataclass
class Record(Serializable):
    date: datetime.date
    properties: list[Property] = field(default_factory=list)
    summary: list[str] = field(default_factory=list)
    entries: list[Entry] = field(default_factory=list)
    tags: list[str] = field(default_factory=list)

    def total_time(self):
        return sum(e.time.total_minutes for e in self.entries)

    def serialize(self) -> str:
        builder = [self.date.strftime('%Y-%m-%d')]

        if len(self.properties) > 0:
            builder.append(
                f'({",".join(p.serialize() for p in self.properties)})')

        builder.append('\n')

        for line in self.summary:
            builder.append(line)
            builder.append('\n')

        for e in self.entries:
            builder.append(f'    {e.serialize()}\n')

        return ''.join(builder)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['klogpy']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.1.2,<8.0.0', 'colorama>=0.4.4,<0.5.0', 'parsy>=1.3,<2.0']

entry_points = \
{'console_scripts': ['klogpy = klogpy.cli:klg']}

setup_kwargs = {
    'name': 'klogpy',
    'version': '0.1.1',
    'description': 'Python implementation of klog format for time tracking',
    'long_description': None,
    'author': 'Nick Yu',
    'author_email': 'nickyu42@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import pickle
from dataclasses import dataclass, field
from pathlib import Path
from typing import Optional

from klogpy import __version__
from klogpy.syntax import Record

CONFIG_DIR = Path.home() / '.klogger'
KLOG_FILE = CONFIG_DIR / 'time.klog'
RECORD_STORE = CONFIG_DIR / 'records.pickle'


@dataclass
class RecordStore:
    VERSION = __version__
    records: list[Record] = field(default_factory=list)

    # TODO: sensible value on init
    current_record_index: Optional[int] = field(default=None)

    @property
    def current_record(self) -> Optional[Record]:
        if self.current_record_index is None:
            return None

        return self.records[self.current_record_index]

    def commit(self):
        """Add changes to this object, to the record store. """
        with RECORD_STORE.open('wb') as f:
            pickle.dump(self, f)

    def write_selected(self) -> Record:
        """
        Write the currently selected record to the Klog file.

        :return: The written record.
        """
        current_record = self.current_record
        if current_record is None:
            raise ValueError('The current record store is empty')

        with KLOG_FILE.open('a') as f:
            f.write(current_record.serialize())
            f.write('\n')

        return current_record

    def push_record(self, record: Record):
        self.records.append(record)

        if self.current_record_index is None:
            self.current_record_index = 0
            return

        self.current_record_index += 1

    def pop_record(self, index: int):
        if index < 0 or index >= len(self.records):
            raise IndexError(f'There is no record at index {index}')

        self.records.pop(index)

        if len(self.records) == 0:
            self.current_record_index = None
        else:
            self.current_record_index -= 1


def get_local_config(should_create: bool = False) -> tuple[bool, Optional[RecordStore]]:
    created = False
    entry_store = None

    if not CONFIG_DIR.exists() and should_create:
        CONFIG_DIR.mkdir()
        RECORD_STORE.touch()
        KLOG_FILE.touch()
        entry_store = RecordStore()

        with RECORD_STORE.open('wb') as f:
            pickle.dump(entry_store, f)

        created = True
    elif CONFIG_DIR.exists() and RECORD_STORE.exists():
        with RECORD_STORE.open('rb') as f:
            entry_store = pickle.load(f)

    return created, entry_store
